from abc import ABC, abstractmethod
from .weights import Weights
import numpy as np
from numpy.typing import ArrayLike


class Layer(ABC):
    """
    this is an abstract class and can only be used indirectly through inherited classes
    """
    __slots__ = ['name', 'mode', 'layerID']
    id = 0

    def __init__(self) -> None:
        self.name = self.__class__.__name__
        self.mode = ''
        self.layerID = Layer.id
        Layer.id += 1

    @property
    def qualifiedName(self) -> tuple:
        return self.__class__.__module__, self.__class__.__name__

    @abstractmethod
    def forward(self, input: ArrayLike) -> np.ndarray:
        """
        it's an abstract method, thus forcing the coder to implement it in daughter classes
        """
        pass

    def __call__(self, *args: ArrayLike) -> np.ndarray:
        """
        this is used to make layers behave more like functions
        """
        return self.forward(*args)

    @abstractmethod
    def backward(self, gradient: ArrayLike) -> np.ndarray:
        """
        it's an abstract method, thus forcing the coder to implement it in daughter classes
        """
        pass

    def train(self) -> None:
        """
        used to put layer in to training mode
        meaning unfreezes parameters
        """
        self.mode = 'train'

    def eval(self) -> None:
        """
        used to put layer in to evaluation mode
        meaning freezes parameters
        """
        self.mode = 'eval'

    def __str__(self) -> str:
        """
        used for print the layer in a human readable manner
        """
        return self.name
