import numpy as np
from abc import ABC, abstractmethod
from .impurityMeasure import ImpurityMeasure
from concurrent.futures import ThreadPoolExecutor, as_completed
from copy import copy
import threading


class SplitAlgorithm(ABC):
    """
    this is a base class for implementing different algorithms
    for splitting features in data sets and finding thresholds
    """
    def __init__(self, featurePercentile: int = None, threads: int = 4) -> None:
        self.name = self.__class__.__name__
        self.featurePercentile = featurePercentile
        self._feature = None
        self._threshold = None
        self.lock = threading.Lock()
        self.threads = threads

    def __call__(self, data: np.ndarray, targets: np.ndarray, weights: np.ndarray = None) -> tuple[float, int, float]:
        """
        making it a bit more convenient to use this class
        """
        return self.findBestSplit(data, targets)

    def setImpurityMeasure(self, impurityMeasure: ImpurityMeasure) -> None:
        """
        the optimal split depends in the impurity measure
        that is why the algorithm needs to have one
        """
        if isinstance(impurityMeasure, ImpurityMeasure):
            self._impurity = impurityMeasure
        else:
            raise TypeError("impurity measure is not of class ImpurityMeasure")

    def getPercentiles(self, features: np.ndarray) -> np.ndarray:
        """
        Get percentiles of feature values
        """
        if self.featurePercentile is None:
            return features
        percentiles = np.linspace(0, 100, self.featurePercentile+2)[1:-1]  # Exclude 0 and 100
        return np.percentile(features, percentiles)

    @abstractmethod
    def _calculateInfoGain(self, feat: int, data: np.ndarray, targets: np.ndarray, weights: np.ndarray = None) -> None:
        pass

    def findBestSplit(self, data: np.ndarray, targets: np.ndarray, weights: np.ndarray = None) -> tuple[float, int, float]:
        self.maxInfoGain = copy(self._maxInfoGain)
        self.feature = copy(self._feature)
        self.threshold = copy(self._threshold)

        # Create a thread pool with a maximum of 4 threads
        with ThreadPoolExecutor(max_workers=self.threads) as executor:
            # Submit each feature to the thread pool for processing
            futures = [executor.submit(self._calculateInfoGain, feat, data, targets, weights) for feat in range(data.shape[1])]

            for future in as_completed(futures):
                future.result()  # This will re-raise any exceptions that occurred in the thread

        return self.feature, self.threshold


class CART(SplitAlgorithm):
    """
    CART (Classification and Regression Trees)
    this is the most robust algorithm
    also suitable for most problems
    """
    def __init__(self, featurePercentile: int = None, threads: int = 4) -> None:
        super().__init__(featurePercentile, threads)
        self._maxInfoGain = np.inf

    def _calculateInfoGain(self, feat: int, data: np.ndarray, targets: np.ndarray, weights: np.ndarray = None) -> None:
        line = data[:, feat]
        splits = self.getPercentiles(line)
        splits = np.unique(splits)
        allSplitsMask = line[:, None] <= splits
        leftCounts = np.sum(allSplitsMask, axis=0)
        rightCounts = len(data) - leftCounts
        validSplits = (leftCounts > 0) & (rightCounts > 0)

        if not np.any(validSplits):
            return

        leftTargets = [targets[splitMask] for splitMask in allSplitsMask.T]
        rightTargets = [targets[np.logical_not(splitMask)] for splitMask in allSplitsMask.T]

        leftImpurity = np.array([(leftCounts[i] / len(data)) * self._impurity(leftTargets[i], weights) for i in range(len(splits))])
        rightImpurity = np.array([(rightCounts[i] / len(data)) * self._impurity(rightTargets[i], weights) for i in range(len(splits))])

        infoGain = leftImpurity + rightImpurity
        infoGain = infoGain[validSplits]
        splits = splits[validSplits]

        minInfoGainIndex = np.argmin(infoGain)
        currentInfoGain = infoGain[minInfoGainIndex]

        with self.lock:
            if currentInfoGain < self.maxInfoGain:
                self.maxInfoGain = currentInfoGain
                self.feature = feat
                self.threshold = splits[minInfoGainIndex]


class ID3(SplitAlgorithm):
    """
    ID3 (Iterative Dichotomiser 3)
    this is used for classification tasks
    less robust than CART
    """
    def __init__(self, featurePercentile: int = None, threads: int = 4) -> None:
        super().__init__(featurePercentile, threads)
        self._maxInfoGain = -np.inf

    def _calculateInfoGain(self, feat: int, data: np.ndarray, targets: np.ndarray, weights: np.ndarray = None) -> None:
        line = data[:, feat]
        splits = self.getPercentiles(line)
        splits = np.unique(splits)
        allSplitsMask = line[:, None] <= splits
        leftCounts = np.sum(allSplitsMask, axis=0)
        rightCounts = len(data) - leftCounts
        validSplits = (leftCounts > 0) & (rightCounts > 0)

        if not np.any(validSplits):
            return

        leftTargets = [targets[splitMask] for splitMask in allSplitsMask.T]
        rightTargets = [targets[np.logical_not(splitMask)] for splitMask in allSplitsMask.T]

        leftImpurity = np.array([(leftCounts[i] / len(data)) * self._impurity(leftTargets[i], weights) for i in range(len(splits))])
        rightImpurity = np.array([(rightCounts[i] / len(data)) * self._impurity(rightTargets[i], weights) for i in range(len(splits))])

        leftImpurityContrib = leftCounts / len(data) * leftImpurity
        rightImpurityContrib = rightCounts / len(data) * rightImpurity

        # Calculate information gain using the ID3 algorithm
        infoGain = self._impurity(targets) - leftImpurityContrib - rightImpurityContrib
        maxIndex = np.argmax(infoGain)

        with self.lock:
            if infoGain[maxIndex] > self.maxInfoGain:
                self.maxInfoGain = infoGain[maxIndex]
                self.feature = feat
                self.threshold = splits[maxIndex]


class C45(SplitAlgorithm):
    """
    C4.5 is an extended version of ID3
    """
    def __init__(self, featurePercentile: int = None, threads: int = 4, epsilon: float = 1e-10) -> None:
        super().__init__(featurePercentile, threads)
        self.epsilon = epsilon # stability constant
        self._maxInfoGain = -np.inf

    def _calculateInfoGain(self, feat: int, data: np.ndarray, targets: np.ndarray, weights: np.ndarray = None) -> None:
        line = data[:, feat]
        splits = self.getPercentiles(line)
        splits = np.unique(splits)
        allSplitsMask = line[:, None] <= splits
        leftCounts = np.sum(allSplitsMask, axis=0)
        rightCounts = len(data) - leftCounts
        validSplits = (leftCounts > 0) & (rightCounts > 0)

        if not np.any(validSplits):
            return

        leftTargets = [targets[splitMask] for splitMask in allSplitsMask.T]
        rightTargets = [targets[np.logical_not(splitMask)] for splitMask in allSplitsMask.T]

        leftImpurity = np.array([(leftCounts[i] / len(data)) * self._impurity(leftTargets[i], weights) for i in range(len(splits))])
        rightImpurity = np.array([(rightCounts[i] / len(data)) * self._impurity(rightTargets[i], weights) for i in range(len(splits))])

        leftImpurityContrib = leftCounts / len(data) * leftImpurity
        rightImpurityContrib = rightCounts / len(data) * rightImpurity

        # Calculate information gain using the C45 algorithm
        infoGain = self._impurity(targets) - leftImpurityContrib - rightImpurityContrib

        # Calculate split info
        left = (leftCounts / len(data))
        right = (rightCounts / len(data))
        splitInfo = - (left * np.log2(left + self.epsilon)) - (right * np.log2(right + self.epsilon))
        splitInfo[splitInfo == 0] += self.epsilon  # avoid dividing by zero

        # Calculate information gain ratio
        infoGainRatio = infoGain / splitInfo

        maxIndex = np.argmax(infoGainRatio)

        with self.lock:
            if infoGainRatio[maxIndex] > self.maxInfoGain:
                self.maxInfoGain = infoGainRatio[maxIndex]
                self.feature = feat
                self.threshold = splits[maxIndex]
